"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, ON, PR)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from subprocess import run, PIPE
from abc import abstractmethod

import numpy as np

from ase.io import read
from ase.io.lammpsdata import write_lammps_data
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.data import chemical_symbols, atomic_masses

from .state import StateManager
from ..core.manager import Manager
from ..utilities import get_elements_Z_and_masses
from ..utilities.io_lammps import (LammpsInput,
                                   EmptyLammpsBlockInput,
                                   LammpsBlockInput,
                                   get_lammps_command)


class BaseLammpsState(StateManager):
    """
    Base class to perform simulations with LAMMPS.
    """
    def __init__(self, nsteps, nsteps_eq, logfile, trajfile, loginterval=50,
                 lammpsfname=None, blocks=None, neti=False, eq_mass_md=False,
                 **kwargs):

        super().__init__(nsteps, nsteps_eq, logfile, trajfile, loginterval,
                         **kwargs)

        self.eq_mass_md = eq_mass_md
        self.ispimd = False
        self.isrestart = False
        self.nbeads = 1  # Dummy nbeads to help

        self.atomsfname = "atoms.in"
        self.lammpsfname = lammpsfname
        if self.lammpsfname is None:
            self.lammpsfname = "lammps_input.in"
        self._myblock = blocks
        # key word to adapt functions from BaseLammsState for NETI
        self.neti = neti

        self.info_dynamics = dict()
        if isinstance(blocks, list):
            self._myblock = blocks[0]
            if len(blocks) != 1:
                for block in blocks[1:]:
                    self._myblock.extend(block)

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def run_dynamics(self,
                     supercell,
                     pair_style,
                     pair_coeff,
                     model_post=None,
                     atom_style="atomic",
                     eq=False,
                     elements=None):
        """
        Function to run the dynamics
        """
        atoms = supercell.copy()

        if elements is None:
            el, Z, masses, charges = get_elements_Z_and_masses(atoms)
            elements = el

        initial_charges = atoms.get_initial_charges()

        blocks = self._get_block_inputs(atoms, pair_style, pair_coeff,
                                        model_post, atom_style, eq, elements)
        if self.neti is False:
            txt = "Lammps input to run MlMD created by MLACS"
            lmp_input = LammpsInput(txt)
        else:
            txt = "Lammps input to run a NETI created by MLACS"
            lmp_input = LammpsInput(txt)
        for block in blocks:
            lmp_input(block.name, block)

        with open(self.subsubdir / self.lammpsfname, "w") as fd:
            fd.write(str(lmp_input))

        self._write_lammps_atoms(atoms, atom_style, elements)

        lmp_cmd = self._get_lammps_command()
        lmp_handle = run(lmp_cmd,
                         shell=True,
                         cwd=str(self.subsubdir),
                         stderr=PIPE)

        if lmp_handle.returncode != 0:
            msg = "LAMMPS stopped with the exit code \n" + \
                  f"{lmp_handle.stderr.decode()}"
            raise RuntimeError(msg)
        if self.neti is False:
            atoms = self._get_atoms_results(initial_charges)

        # Set the info of atoms
        atoms.info['info_state'] = self.info_dynamics

        return atoms.copy()

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def _write_lammps_atoms(self, atoms, atom_style, elements):
        """

        """
        write_lammps_data(str(self.subsubdir / self.atomsfname),
                          atoms,
                          velocities=True,
                          atom_style=atom_style,
                          specorder=elements.tolist())


# ========================================================================== #
    def _get_block_inputs(self, atoms, pair_style, pair_coeff, model_post,
                          atom_style, eq, elements):
        """

        """
        masses = [atomic_masses[chemical_symbols.index(element)]
                  for element in elements]

        pbc = atoms.get_pbc()
        if self.eq_mass_md:
            masses = np.ones(np.shape(masses))

        blocks = []
        blocks.append(self._get_block_init(atom_style, pbc, elements, masses))
        blocks.append(self._get_block_interactions(pair_style, pair_coeff,
                                                   model_post, atom_style,
                                                   atoms))
        blocks.append(self._get_block_thermostat(eq))
        if self.logfile is not None:
            blocks.append(self._get_block_log())
        if self.trajfile is not None:
            blocks.append(self._get_block_traj(elements))
        if isinstance(self._get_block_custom(), list):
            blocks.extend(self._get_block_custom())
        else:
            blocks.append(self._get_block_custom())
        blocks.append(self._get_block_run(eq))
        if self.neti is False:
            blocks.append(self._get_block_lastdump(elements, eq))
        return blocks

# ========================================================================== #
    def _get_block_init(self, atom_style, pbc, el, masses):
        """

        """
        pbc = "{0} {1} {2}".format(*tuple("sp"[int(x)] for x in pbc))

        block = LammpsBlockInput("init", "Initialization")
        block("units", "units metal")
        block("boundary", f"boundary {pbc}")
        block("atom_style", f"atom_style {atom_style}")
        block("read_data", f"read_data {self.atomsfname}")
        for i, mass in enumerate(masses):
            block(f"mass{i}", f"mass {i+1}  {mass}")
        for iel, e in enumerate(el):
            block("group", f"group {e} type {iel+1}")
        return block

# ========================================================================== #
    def _get_block_run(self, eq):
        """

        """
        if eq:
            nsteps = self.nsteps_eq
        else:
            nsteps = self.nsteps
        block = LammpsBlockInput("run")
        block("run", f"run {nsteps}")
        return block

# ========================================================================== #
    def _get_block_interactions(self, pair_style, pair_coeff, model_post,
                                atom_style, atoms):
        """

        """
        # Write lammps input
        block = LammpsBlockInput("interaction", "Interaction")
        block("pair_style", f"pair_style {pair_style}")
        for i, pair in enumerate(pair_coeff):
            block(f"pair_coeff{i}", f"pair_coeff {pair}")
        if model_post is not None:
            for i, model in enumerate(model_post):
                block(f"model{i}", f"{model}")
        return block

# ========================================================================== #
    def _get_block_thermostat(self, eq):
        return None

# ========================================================================== #
    def _get_block_log(self):
        """

        """
        block = LammpsBlockInput("log", "Logging")
        variables = ["t equal step", "mytemp equal temp",
                     "mype equal pe", "myke equal ke", "myetot equal etotal",
                     "mypress equal press/10000", "vol equal (lx*ly*lz)"]
        for i, var in enumerate(variables):
            block(f"variable{i}", f"variable {var}")
        txt = f"fix mylog all print {self.loginterval} " + \
              '"$t ${mytemp} ${vol} ${myetot} ${mype} ${myke} ${mypress}" ' + \
              f"append {self.logfile} title " + \
              '"# Step Temp Vol Etot Epot Ekin Press"'
        block("fix", txt)
        return block

# ========================================================================== #
    def _get_block_lastdump(self, el, eq):
        """

        """
        block = LammpsBlockInput("lastdump", "Dump last configuration")
        txt = "dump last all custom 1 configurations.out " + \
              "id type xu yu zu vx vy vz fx fy fz element"
        block("dump", txt)
        txt = "dump_modify last element " + " ".join([p for p in el])
        block("dump_modify1", txt)
        block("run_dump", "run 0")
        return block

# ========================================================================== #
    def _get_block_traj(self, el):
        """

        """
        block = LammpsBlockInput("traj", "Dumping trajectory")
        txt = f"dump dum1 all custom {self.loginterval} {self.trajfile} " + \
              "id type xu yu zu vx vy vz fx fy fz "
        txt += "element"
        block("dump", txt)
        block("dump_modify1", "dump_modify dum1 append yes")
        txt = "dump_modify dum1 element " + " ".join([p for p in el])
        block("dump_modify2", txt)
        return block

# ========================================================================== #
    def _get_block_custom(self):
        """

        """
        if isinstance(self._myblock, LammpsBlockInput):
            return self._myblock
        else:
            return EmptyLammpsBlockInput("empty_custom")

# ========================================================================== #
    def _get_atoms_results(self, initial_charges):
        """

        """
        atoms = read(self.subsubdir / "configurations.out")
        if initial_charges is not None:
            atoms.set_initial_charges(initial_charges)
        return atoms

# ========================================================================== #
    def _get_lammps_command(self):
        '''
        Function to load the bash command to run LAMMPS
        '''
        cmd = get_lammps_command()
        return f"{cmd} -in {self.lammpsfname} -sc out.lmp"

# ========================================================================== #
    def initialize_momenta(self, atoms):
        """

        """
        pass

# ========================================================================== #
    @abstractmethod
    def log_recap_state(self):
        pass


# ========================================================================== #
# ========================================================================== #
class LammpsState(BaseLammpsState):
    """
    Class to perform NVT or NPT simulations with LAMMPS.

    Parameters
    ----------
    temperature: :class:`float`
        Temperature of the simulation, in Kelvin.

    pressure: :class:`float` or ``None`` (optional)
        Pressure of the simulation, in GPa.
        If ``None``, no barostat is applied and
        the simulation is in the NVT ensemble. Default ``None``

    t_stop: :class:`float` or ``None`` (optional)
        When this input is not ``None``, the temperature of
        the molecular dynamics simulations is randomly chosen
        in a range between `temperature` and `t_stop`.
        Default ``None``

    p_stop: :class:`float` or ``None`` (optional)
        When this input is not ``None``, the pressure of
        the molecular dynamics simulations is randomly chosen
        in a range between `pressure` and `p_stop`.
        Naturally, the `pressure` input has to be set.
        Default ``None``

    damp: :class:`float` or ``None`` (optional)
        The damping value for the thermostat.
        The default gives a sensible value of a hundred times the
        timestep.

    langevin: :class:`Bool` (optional)
        If ``True``, a Langevin thermostat is used for the thermostat.
        Default ``True``

    gjf: ``no`` or ``vfull`` or ``vhalf`` (optional)
        Whether to use the Gronbech-Jensen/Farago integrator
        for the Langevin dynamics. Only apply if langevin is ``True``.
        Default ``vhalf``.

    qtb: :clas::`Bool` (optional)
        Whether to use a quantum thermal bath to approximate quantum effects.
        If True, it override the langevin and gjf inputs.
        Default False

    fd: :class:`float` (optional)
        The frequency cutoff for the qtb thermostat. Should be around
        2~3 times the Debye frequency. In THz.
        Default 200 THz.

    n_f: :class:`int` (optional)
        Frequency grid size for the qtb thermostat.
        Default 100.

    pdamp: :class:`float` or ``None`` (optional)
        Damping parameter for the barostat.
        If ``None``, apply a damping parameter of
        1000 times the timestep of the simulation. Default ``None``

    ptype: ``iso`` or ``aniso`` (optional)
        Handle the type of pressure applied. Default ``iso``

    twodimensional: :class:`bool` (optional)
        If set to ``True`` and pressure is not ``None``, set the pressure
        only on the x and y axis. Pressure along `x` and `y` axis can
        still be coupled by setting ``ptype`` to `iso`.
        default ``False``

    dt : :class:`float` (optional)
        Timestep, in fs. Default ``1.5`` fs.

    fixcm : :class:`Bool` (optional)
        Fix position and momentum center of mass. Default ``True``.

    rng : RNG object (optional)
        Rng object to be used with the Langevin thermostat.
        Default correspond to :class:`numpy.random.default_rng()`

    init_momenta : :class:`numpy.ndarray` (optional)
        Gives the (Nat, 3) shaped momenta array that will be used
        to initialize momenta when using
        the `initialize_momenta` function.
        If the default ``None`` is set, momenta are initialized with a
        Maxwell Boltzmann distribution.

    nsteps : :class:`int` (optional)
        Number of MLMD steps for production runs. Default ``1000`` steps.

    nsteps_eq : :class:`int` (optional)
        Number of MLMD steps for equilibration runs. Default ``100`` steps.

    eq_mass_md : :class:`Bool` (optional)
        If all atoms have the same mass for the MD. Default ``False``
        If True, make sure your MLIP is also correctly parametrized

    logfile : :class:`str` (optional)
        Name of the file for logging the MLMD trajectory.
        If ``None``, no log file is created. Default ``None``.

    trajfile : :class:`str` (optional)
        Name of the file for saving the MLMD trajectory.
        If ``None``, no traj file is created. Default ``None``.

    loginterval : :class:`int` (optional)
        Number of steps between MLMD logging. Default ``50``.

    blocks : :class:`LammpsBlockInput` or :class:`list` (optional)
        Custom block input class. Can be a list of blocks.
        If ``None``, nothing is added in the input. Default ``None``.

    Examples
    --------

    >>> from mlacs.state import LammpsState
    >>>
    >>> state = LammpsState(temperature=300, pressure=None) #NVT
    >>> state = LammpsState(temperature=300, pressure=0)    #NPT
    >>> state.run_dynamics(atoms, mlip.pair_style, mlip.pair_coeff)
    """
    def __init__(self,
                 temperature,
                 pressure=None,
                 t_stop=None,
                 p_stop=None,
                 damp=None,
                 langevin=True,
                 gjf="vhalf",
                 qtb=False,
                 fd=200,
                 n_f=100,
                 pdamp=None,
                 ptype="iso",
                 twodimensional=False,
                 dt=1.5,
                 fixcm=True,
                 rng=None,
                 init_momenta=None,
                 nsteps=1000,
                 nsteps_eq=100,
                 eq_mass_md=False,
                 logfile=None,
                 trajfile=None,
                 loginterval=50,
                 blocks=None,
                 folder='Trajectory',
                 **kwargs):

        kwargs.setdefault('prefix', folder)  # To keep previous behaviour

        super().__init__(nsteps, nsteps_eq, logfile, trajfile,
                         loginterval=loginterval, blocks=blocks,
                         eq_mass_md=eq_mass_md, folder=folder, **kwargs)

        self.temperature = temperature
        self.pressure = pressure
        self.t_stop = t_stop
        self.p_stop = p_stop
        self.damp = damp
        self.langevin = langevin
        self.gjf = gjf
        self.qtb = qtb
        self.fd = fd
        self.n_f = n_f
        self.pdamp = pdamp
        self.ptype = ptype
        self.dt = dt
        self.fixcm = fixcm
        self.rng = rng
        self.init_momenta = init_momenta
        self.twodimensional = twodimensional

        if self.rng is None:
            self.rng = np.random.default_rng()
        if self.damp is None:
            self.damp = "$(100*dt)"
        if self.pdamp is None:
            self.pdamp = "$(1000*dt)"

        if self.p_stop is not None:
            if self.pressure is None:
                msg = "You need to put a pressure with p_stop"
                raise ValueError(msg)

        self._make_info_dynamics()

# ========================================================================== #
    def _get_block_thermostat(self, eq):
        """

        """
        if self.t_stop is None:
            temp = self.temperature
        else:
            tmp_temp = np.sort([self.temperature, self.t_stop])
            if eq:
                temp = np.max(tmp_temp)
            else:
                temp = self.rng.uniform(*tmp_temp)
            self.info_dynamics["temperature"] = temp
        if self.p_stop is None:
            press = self.pressure
        else:
            press = self.rng.uniform(self.pressure, self.p_stop)
            self.info_dynamics["pressure"] = press
        if self.qtb:
            qtbseed = self.rng.integers(1, 99999999)
        if self.langevin:
            langevinseed = self.rng.integers(1, 9999999)

        block = LammpsBlockInput("thermostat", "Thermostat")
        block("timestep", f"timestep {self.dt / 1000}")

        # If we are using Langevin, we want to remove the random part
        # of the forces
        if self.langevin:
            block("rmv_langevin", "fix ff all store/force")

        if self.pressure is None:
            if self.qtb:
                block("nve", "fix f1 all nve")
                txt = f"fix f2 all qtb temp {temp} damp {self.damp} " + \
                      f"f_max {self.fd} N_f {self.n_f} seed {qtbseed}"
                block("qtb", txt)
            elif self.langevin:
                txt = f"fix f1 all langevin {temp} {temp} {self.damp} " + \
                      f"{langevinseed} gjf {self.gjf} zero yes"
                block("langevin", txt)
                block("nve", "fix f2 all nve")
            else:
                block("nvt", f"fix f1 all nvt temp {temp} {temp} {self.damp}")
        else:
            if self.qtb:
                txt = f"fix f1 all nph {self.ptype} " + \
                      f"{press*10000} {press*10000} {self.pdamp}"
                block("nph", txt)
                txt = f"fix f1 all qtb temp {temp} damp {self.damp}" + \
                      f"f_max {self.fd} N_f {self.n_f} seed {qtbseed}"
                block("qtb", txt)
            elif self.langevin:
                txt = f"fix f1 all langevin {temp} {temp} {self.damp} " + \
                      f"{langevinseed} gjf {self.gjf} zero yes"
                block("langevin", txt)
                ptxt = f"{press*10000} {press*10000} {self.pdamp}"
                txt = "fix f2 all nph "
                if self.twodimensional:
                    txt += f"x {ptxt} y {ptxt} "
                    if self.ptype == "iso":
                        txt += "couple xy "
                else:
                    txt += f"{self.ptype} {ptxt}"
                block("nph", txt)
            else:
                txt = f"fix f1 all npt temp {temp} {temp} {self.damp} " + \
                      f"{self.ptype} {press*10000} {press*10000} {self.pdamp}"
                block("npt", txt)
        if self.fixcm:
            block("cm", "fix fcm all recenter INIT INIT INIT")
        return block

# ========================================================================== #
    def _make_info_dynamics(self):

        # NVT, NPT, no (uVT, uPT, NVE) yet
        ensemble = ["X", "X", "X"]

        # N or mu
        ensemble[0] = "N"

        # V or P
        ensemble[1] = "V"
        pressure = None
        if self.pressure is not None:
            ensemble[1] = "P"
            if self.p_stop is None:
                pressure = self.pressure

        # T or E
        if self.temperature is None:  # NVE
            raise NotImplementedError
        ensemble[2] = "T"
        temperature = self.temperature

        # NVT, NPT, no (uVT, uPT, NVE) yet
        self.ensemble = ''.join(ensemble)
        self.info_dynamics = dict(ensemble=self.ensemble,
                                  temperature=temperature,
                                  pressure=pressure)

# ========================================================================== #
    def _get_block_traj(self, el):
        """

        """
        block = LammpsBlockInput("traj", "Dumping trajectory")
        txt = f"dump dum1 all custom {self.loginterval} {self.trajfile} " + \
              "id type xu yu zu vx vy vz fx fy fz "
        if self.langevin:
            txt += "f_ff[1] f_ff[2] f_ff[3] "
        txt += "element"
        block("dump", txt)
        block("dump_modify1", "dump_modify dum1 append yes")
        txt = "dump_modify dum1 element " + " ".join([p for p in el])
        block("dump_modify2", txt)
        return block

# ========================================================================== #
    def initialize_momenta(self, atoms):
        """
        """
        if self.init_momenta is None:
            MaxwellBoltzmannDistribution(atoms,
                                         temperature_K=self.temperature,
                                         rng=self.rng)
        else:
            atoms.set_momenta(self.init_momenta)

        atoms.info['info_state'] = self.info_dynamics

# ========================================================================== #
    def log_recap_state(self):
        """
        Function to return a string describing the state for the log
        """
        damp = self.damp
        if damp is None:
            damp = 100 * self.dt
        pdamp = self.pdamp
        if pdamp is None:
            pdamp = 1000 * self.dt

        if self.temperature is None and self.pressure is None:
            msg = "Geometry optimization as implemented in LAMMPS\n"
        elif self.pressure is None:
            msg = "NVT dynamics as implemented in LAMMPS\n"
        else:
            msg = "NPT dynamics as implemented in LAMMPS\n"
        msg += f"Temperature (in Kelvin)                 {self.temperature}\n"
        if self.langevin:
            msg += "A Langevin thermostat is used\n"
        if self.pressure is not None:
            msg += f"Pressure (GPa)                          {self.pressure}\n"
        msg += f"Number of MLMD equilibration steps :    {self.nsteps_eq}\n"
        msg += f"Number of MLMD production steps :       {self.nsteps}\n"
        msg += f"Timestep (in fs) :                      {self.dt}\n"
        if self.temperature is not None:
            msg += f"Themostat damping parameter (in fs) :   {damp}\n"
            if self.pressure is not None:
                msg += f"Barostat damping parameter (in fs) :    {pdamp}\n"
        msg += "\n"
        return msg



import os
from setuptools import setup, find_packages

with open("README.md") as f:
    long_description = f.read()

versionfile = os.path.join('mlacs', 'version.py')
with open(versionfile) as f:
    code = compile(f.read(), versionfile, 'exec')
    exec(code)

install_requires = ["numpy>=1.17.0",
                    "ase>=3.22"]

extra_requires = ["icet>=1.4",
                  "scikit-learn",
                  "pymbar"
                  "cython",
                  "netCDF4==1.2.2"]


if __name__ == "__main__":
    setup(
          name="mlacs",
          version=__version__,
          author="Aloïs Castellano",
          author_email="alois.castellano@uliege.be",
          packages=find_packages(),
          entry_points={"console_scripts": ["mlacs=mlacs.cli.main:main"]},
          description="Machine-Learning Assisted Canonical Sampling",
          license_files=("LICENSE.txt",),
          long_description=long_description,
          install_requires=install_requires,
          extra_requires=extra_requires
         )



"""
// Copyright (C) 2022-2024 MLACS group (AC, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import sys
import numpy as np

from . import MlipManager
from .weights import UniformWeight
from ..utilities import compute_correlation, create_link

from ase.units import GPa
from ase import Atoms
from pathlib import Path


# ========================================================================== #
# ========================================================================== #
class TensorpotPotential(MlipManager):
    """
    Potential that use Tensorpotential to minimize a cost function.
    For now, only works with AceDescriptor

    Parameters
    ----------

    descriptor: :class:`Descriptor`
        The descriptor used in the model.

    weight: :class:`WeightingPolicy`
        Weight used for the fitting and calculation of properties.
        Default :class:`None`

    Examples
    --------

    >>> from ase.io import read
    >>> confs = read('Trajectory.traj', index=':')
    >>>
    >>> from mlacs.mlip import AceDescriptor, TensorpotPotential
    >>> desc = AceDescriptor(confs[0], free_at_e={'Cu': 0}, rcut=rcut)
    >>> mlip = TensorpotPotential(desc, folder="ACE")
    >>>
    >>> mlip.update_matrices(confs)
    >>> mlip.train_mlip()

    """
    def __init__(self,
                 descriptor,
                 weight=None,
                 **kwargs):
        if weight is None:
            weight = UniformWeight(energy_coefficient=1.0,
                                   forces_coefficient=1.0,
                                   stress_coefficient=0.0)

        MlipManager.__init__(self,
                             descriptor=descriptor,
                             weight=weight,
                             **kwargs)
        self.natoms = []
        self.nconfs = 0

        if self.weight.stress_coefficient != 0:
            raise ValueError("Tensorpotential can't fit on stress")

        self.atoms = []
        self.coefficients = []

        # Remove the old pckl.gzip file to avoid having configuration twice
        old_df = self.descriptor.subsubdir / self.descriptor.db_fn
        if old_df.exists():
            old_df.unlink()

# ========================================================================== #
    def update_matrices(self, atoms):
        """
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        self.weight.update_database(atoms)
        self.nconfs += len(atoms)
        self.natoms.extend([len(_) for _ in atoms])
        self.atoms.extend(atoms)

# ========================================================================== #
    def train_mlip(self):
        """
        """
        if self.descriptor.log is None:
            self.descriptor.redirect_logger()

        self.weight.workdir = self.workdir
        self.weight.folder = self.folder
        self.weight.subfolder = self.subfolder
        self.descriptor.workdir = self.workdir
        self.descriptor.folder = self.folder
        self.descriptor.subfolder = self.subfolder

        W = self.weight.get_weights()
        fitting_log = self.descriptor.log.handlers[0].baseFilename
        with open(fitting_log, "a") as f:
            sys.stdout = f
            coef_fn = self.descriptor.fit(weights=W, atoms=self.atoms)
            full_fn = self.workdir / self.folder / self.subfolder / coef_fn
            self.coefficients = full_fn
            sys.stdout = sys.__stdout__

        msg = "Number of configurations for training: " + \
              f"{self.nconfs}\n"
        msg += "Number of atomic environments for training: " + \
               f"{sum(self.natoms)}\n"

        tmp_msg, weight_fn = self.weight.compute_weight(
            self.coefficients,
            self.predict)

        msg += tmp_msg
        msg += self.compute_tests()

        create_link(self.subsubdir / weight_fn, self.subdir/weight_fn)
        create_link(self.subsubdir / coef_fn, self.subdir/coef_fn)

        if self.log:
            self.log.write(msg)

# ========================================================================== #
    def read_parent_mlip(self, traj):
        """
        Get a list of all the mlip that have generated a conf in traj
        and get the coefficients of all these mlip
        """
        parent_mlip = []
        mlip_coef = []
        directory = self.descriptor.subdir

        # Make the MBAR variable Nk and mlip_coef
        for state in traj:
            for conf in state:
                if "parent_mlip" not in conf.info:  # Initial or training
                    continue
                else:  # A traj
                    model = conf.info['parent_mlip']
                    directory = Path(model)
                    if not directory.exists:
                        err = "Some parent MLIP are missing. "
                        err += "Rerun MLACS with DatabaseCalculator and "
                        err += "OtfMlacs.keep_tmp_files=True on your traj"
                        raise FileNotFoundError(err)
                    if model not in parent_mlip:  # New state
                        parent_mlip.append(model)
                        coef = self.descriptor.get_mlip_file(folder=model)
                        mlip_coef.append(coef)
        return parent_mlip, np.array(mlip_coef)

# ========================================================================== #
    def next_coefs(self, mlip_coef):
        """
        Update MLACS just like train_mlip, but without actually computing
        the coefficients
        """
        self.weight.subfolder = self.subfolder
        self.descriptor.subfolder = self.subfolder

        self.coefficients = mlip_coef

        self.descriptor.set_restart_coefficient(mlip_coef)
        _, weight_fn = self.weight.compute_weight(mlip_coef,
                                                  self.predict,
                                                  docalc=False)

        if isinstance(mlip_coef, str):
            mlip_coef = Path(mlip_coef)
        create_link(mlip_coef / "ACE.yace", self.subdir/"ACE.yace")

# ========================================================================== #
    def compute_tests(self):
        """
        Compute the weighted RMSE and MAE
        """
        e_mlip, f_mlip, s_mlip = self.predict(atoms=self.atoms)
        f_mlip = np.concatenate([f.flatten() for f in f_mlip])
        s_mlip = np.concatenate(s_mlip)

        e_true = np.array([at.get_potential_energy()/len(at)
                          for at in self.atoms])
        f_true = np.concatenate([at.get_forces().flatten()
                                 for at in self.atoms])
        s_true = np.concatenate([at.get_stress() for at in self.atoms])

        w = None
        if len(self.weight.weight) > 0:
            w = self.weight.weight
        wf = np.array([])
        for i in range(len(w)):
            wf = np.append(wf, np.ones(self.natoms[i]*3)*(w[i]/3))

        res_E = compute_correlation(np.c_[e_true, e_mlip], weight=w)
        res_F = compute_correlation(np.c_[f_true, f_mlip], weight=wf)
        res_S = compute_correlation(np.c_[s_true, s_mlip]/GPa, weight=w)
        r = np.c_[res_E, res_F, res_S]
        self.fit_res = r

        # Information to MLIP-Energy_comparison.dat
        header = f"Weighted rmse: {self.fit_res[0, 0]:.6f} eV/at,    " + \
                 f"Weighted mae: {self.fit_res[1, 0]:.6f} eV/at\n" + \
                 " True Energy           Predicted Energy"
        np.savetxt("MLIP-Energy_comparison.dat",
                   np.c_[e_true, e_mlip],
                   header=header, fmt="%25.20f  %25.20f")
        header = f"Weighted rmse: {self.fit_res[0, 1]:.6f} eV/angs   " + \
                 f"Weighted mae: {self.fit_res[1, 1]:.6f} eV/angs\n" + \
                 " True Forces           Predicted Forces"

        np.savetxt("MLIP-Forces_comparison.dat",
                   np.c_[f_true, f_mlip],
                   header=header, fmt="%25.20f  %25.20f")
        header = f"Weighted rmse: {self.fit_res[0, 2]:.6f} GPa       " + \
                 f"Weighted mae: {self.fit_res[1, 2]:.6f} GPa\n" + \
                 " True Stress           Predicted Stress"
        np.savetxt("MLIP-Stress_comparison.dat",
                   np.c_[s_true, s_mlip] / GPa,
                   header=header, fmt="%25.20f  %25.20f")

        # Message to Mlacs.log
        msg = f"Weighted RMSE Energy    {self.fit_res[0, 0]:.4f} eV/at\n"
        msg += f"Weighted MAE Energy     {self.fit_res[1, 0]:.4f} eV/at\n"
        msg += f"Weighted RMSE Forces    {self.fit_res[0, 1]:.4f} eV/angs\n"
        msg += f"Weighted MAE Forces     {self.fit_res[1, 1]:.4f} eV/angs\n"
        msg += f"Weighted RMSE Stress    {self.fit_res[0, 2]:.4f} GPa\n"
        msg += f"Weighted MAE Stress     {self.fit_res[1, 2]:.4f} GPa\n"
        return msg

# ========================================================================== #
    def get_calculator(self):
        """
        Initialize a ASE calculator from the model
        """
        from .calculator import MlipCalculator
        calc = MlipCalculator(self)
        return calc

# ========================================================================== #
    def predict(self, atoms, coef=None):
        """
        Give the energy forces stress of atoms according to the potential.
        """
        if coef is None:
            coef = self.coefficients
        if isinstance(coef, str):
            coef = Path(coef)
        return self.descriptor.predict(atoms, coef, folder=self.subdir)

# ========================================================================== #
    def __str__(self):
        dname = type(self.descriptor).__name__
        txt = f"Tensorpotential potential with {dname}"
        return txt

# ========================================================================== #
    def __repr__(self):
        txt = "Tensorpotential\n"
        txt += "Parameters:\n"
        txt += "-----------\n"
        txt += f"Weight : {type(self.weight).__name__}"
        txt += "\n"
        txt += "Descriptor used in the potential:\n"
        txt += repr(self.descriptor)
        return txt



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np

from abc import ABC, abstractmethod

from ase.neighborlist import neighbor_list

from ..core import Manager
from ..utilities import get_elements_Z_and_masses


# ========================================================================== #
# ========================================================================== #
class Descriptor(Manager, ABC):
    """
    Base class for descriptors

    Parameters
    ----------
    atoms: :class:`ase.atoms`
        Reference structure, with the elements for the descriptor

    rcut: :class:`float`
        The cutoff for the descriptor
    """
# ========================================================================== #
    def __init__(self, atoms, rcut=5.0, alpha=1.0, prefix='MLIP', **kwargs):

        Manager.__init__(self, prefix=prefix, **kwargs)

        if isinstance(atoms, list):
            self.elements, self.Z, self.masses, self.charges = \
                    [np.array([]) for _ in range(4)]
            for at in atoms:
                el, Z, masses, charges = get_elements_Z_and_masses(at)
                for i in range(len(el)):
                    if el[i] not in self.elements:
                        self.elements = np.append(self.elements, el[i])
                        self.Z = np.append(self.Z, Z[i])
                        self.masses = np.append(self.masses, masses[i])
                        if charges is None:
                            self.charges = np.append(self.charges, 0.)
                        else:
                            self.charges = np.append(self.charges, charges[i])
            if np.allclose(self.charges, 0.0, atol=1e-8):
                self.charges = None
        else:
            self.elements, self.Z, self.masses, self.charges = \
              get_elements_Z_and_masses(atoms)
        self.nel = len(self.elements)
        self.rcut = rcut
        self.welems = np.array(self.Z) / np.sum(self.Z)
        self.alpha = alpha
        self.need_neigh = False

# ========================================================================== #
    def compute_descriptors(self, atoms, forces=True, stress=True):
        desc = []
        for at in atoms:
            # AC : apparently, the at.info for descriptor does not work
            # RB : Tested this, the fix works fined for Trajectory confs but
            #      not for Training. I don't know why ?!
            if "descriptor" in at.info:
                desc.append(at.info['descriptor'])
            else:
                desc.append(self.compute_descriptor(atoms=at,
                                                    forces=forces,
                                                    stress=stress))
        return desc

# ========================================================================== #
    @abstractmethod
    def compute_descriptor(self, atoms, forces=True, stress=True):
        pass

# ========================================================================== #
    def _compute_rij(self, atoms):
        """
        """
        iat, jat, vdist = neighbor_list("ijD", atoms, self.rcut)
        iel = np.array(atoms.get_chemical_symbols())
        iel = np.array([np.where(self.elements == el)[0][0]
                        for el in iel])
        return iat, jat, vdist, iel

# ========================================================================== #
    def _regularization_matrix(self):
        """
        """
        return np.eye(self.ncolumns) * self.alpha


# ========================================================================== #
# ========================================================================== #
class SumDescriptor(Descriptor):
    """
    A class to mix several descriptors together.

    Parameters
    ----------
    args: :class:`list` of :class:`Descriptor`
        A list of all the descriptors to mix.
    """
    def __init__(self, *args):
        # This is wrong for write_mlip. I don't know if there is still
        # a use case for this now that we have DeltaLearningPotential
        raise NotImplementedError("SumDescriptor are not functional")
        self.desc = args
        self.elements = self.desc[0].elements.copy()
        self.rcut = np.max([d.rcut for d in self.desc])
        self.need_neigh = np.any([d.need_neigh for d in self.desc])
        self.ncolumns = np.sum([d.ncolumns for d in self.desc])

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def write_mlip(self, coefficients):
        icol = 0
        for d in self.desc:
            fcol = icol + d.ncolumns
            d.write_mlip(coefficients[icol:fcol])
            icol = fcol

# ========================================================================== #
    def compute_descriptor(self, atoms, forces=True, stress=True):
        pass

# ========================================================================== #
    def _regularization_matrix(self):
        """
        """
        reg = []
        for d in self.desc:
            reg.append(d._regularization_matrix())
        reg = combine_reg(reg)
        return reg

# ========================================================================== #
    def get_pair_style(self):
        pair_style = "hybrid/overlay "
        for d in self.desc:
            pair_style_d = d.get_pair_style()
            pair_style += f"{pair_style_d} "
        return pair_style

# ========================================================================== #
    def get_pair_coeff(self):
        pair_coeff = []
        for d in self.desc:
            pair_style_d, pair_coeff_d = d.get_pair_style_coeff()
            for coeff in pair_coeff_d:
                style = pair_style_d.split()[0]
                co = coeff.split()
                co.insert(2, style)
                pair_coeff.append(" ".join(co))
        return pair_coeff

# ========================================================================== #
    def get_pair_style_coeff(self):
        return self.get_pair_style(), self.get_pair_coeff()

# ========================================================================== #
    def to_dict(self):
        des = []
        for d in self.desc:
            des.append(d.to_dict())
        dct = dict(name="SumDescriptor",
                   descriptor=des)
        return dct

# ========================================================================== #
    @staticmethod
    def from_dict(dct):
        dct.pop("name", None)
        alldesc = []
        for d in dct["descriptor"]:
            name = d.pop("name")
            import mlacs.mlip as tmp
            descclass = getattr(tmp, name)
            desc = descclass.from_dict(d)
            alldesc.append(desc)
        return SumDescriptor(*alldesc)

# ========================================================================== #
    def __str__(self):
        """
        """
        txt = f"Sum descriptor composed of {len(self.desc)} descriptor"
        return txt

# ========================================================================== #
    def __repr__(self):
        """
        """
        txt = "Sum descriptor\n"
        txt += "--------------\n"
        txt += f"Number of descriptor : {len(self.desc)}\n"
        txt += f"Max rcut :             {self.rcut}\n"
        txt += f"Dimension total :      {self.ncolumns}\n"
        for i, d in enumerate(self.desc):
            txt += f"Descriptors {i+1}:\n"
            txt += repr(d)
            txt += "\n"
        return txt


# ========================================================================== #
# ========================================================================== #
class BlankDescriptor(Descriptor):
    """
    A blank descriptor to serve as Dummy for model that compute the
    descriptor AND do the regression
    """
    def __init__(self, atoms):
        Descriptor.__init__(self, atoms)

# ========================================================================== #
    def compute_descriptor(self, atoms, forces=True, stress=True):
        pass

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def write_mlip(self, mlip_coef):
        pass


# ========================================================================== #
def combine_reg(matrices):
    """
    Combine regularization matrices. Adapted from UF3 code
    available on
    https://github.com/uf3/uf3/blob/master/uf3/regression/regularize.py
    """
    sizes = np.array([len(m) for m in matrices])
    nfeat = int(sizes.sum())
    fullmat = np.zeros((nfeat, nfeat))
    origins = np.insert(np.cumsum(sizes), 0, 0)
    for i, mat in enumerate(matrices):
        start = origins[i]
        end = origins[i + 1]
        fullmat[start:end, start:end] = mat
    return fullmat



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from ... import context  # noqa
from mlacs.utilities.io_lammps import (LammpsInput, LammpsBlockInput,
                                       EmptyLammpsBlockInput)


pred1 = """######################
#     test title     #
######################
test_variable 1
test_variable 2
"""


pred2 = """######################
#     test title     #
######################
test_variable 1
test_variable 3
test_variable 2
"""


pred3 = """######################
#     test title     #
######################
test_variable 4
test_variable 1
test_variable 3
test_variable 2
"""


def test_block():
    block_test = LammpsBlockInput("test", "test title")
    block_test.add_variable("test_variable1", "test_variable 1")
    block_test.add_variable("test_variable2", "test_variable 2")

    for pred, ref in zip(pred1.split("\n"), str(block_test).split("\n")):
        assert pred == ref

    block_test.add_variable("test_variable3", "test_variable 3", order=2)
    for pred, ref in zip(pred2.split("\n"), str(block_test).split("\n")):
        assert pred == ref

    block_test.add_variable("test_variable4", "test_variable 4",
                            before="test_variable1")
    for pred, ref in zip(pred3.split("\n"), str(block_test).split("\n")):
        assert pred == ref

    block_test.pop("test_variable4")
    for pred, ref in zip(pred2.split("\n"), str(block_test).split("\n")):
        assert pred == ref

    emptyblock_test = EmptyLammpsBlockInput("empty_test")
    assert str(emptyblock_test) == ""


lmp_in1 = """# preambule

##################
#     block1     #
##################
input 1

##################
#     block2     #
##################
input 2
"""


lmp_in2 = """# preambule

##################
#     block1     #
##################
input 1

##################
#     block3     #
##################
input 3

##################
#     block2     #
##################
input 2
"""


lmp_in3 = """# preambule

##################
#     block4     #
##################
input 4

##################
#     block1     #
##################
input 1

##################
#     block3     #
##################
input 3

##################
#     block2     #
##################
input 2
"""


def test_lammps_input():
    block1 = LammpsBlockInput("block1", "block1")
    block1.add_variable("input1", "input 1")
    block2 = LammpsBlockInput("block2", "block2")
    block2.add_variable("input2", "input 2")

    lmp_in = LammpsInput("preambule")
    lmp_in.add_block("block1", block1)
    lmp_in.add_block("block2", block2)
    for pred, ref in zip(lmp_in1.split("\n"), str(lmp_in).split("\n")):
        assert pred == ref

    block3 = LammpsBlockInput("block3", "block3")
    block3.add_variable("input3", "input 3")

    lmp_in.add_block("block3", block3, order=2)
    for pred, ref in zip(lmp_in2.split("\n"), str(lmp_in).split("\n")):
        assert pred == ref

    block4 = LammpsBlockInput("block4", "block4")
    block4.add_variable("input4", "input 4")

    lmp_in.add_block("block4", block4, before="block1")
    for pred, ref in zip(lmp_in3.split("\n"), str(lmp_in).split("\n")):
        assert pred == ref

    lmp_in.pop("block4")
    for pred, ref in zip(lmp_in2.split("\n"), str(lmp_in).split("\n")):
        assert pred == ref



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from ..core.manager import Manager

from ase.calculators.singlepoint import SinglePointCalculator
from ase.calculators.calculator import CalculatorError


# ========================================================================== #
# ========================================================================== #
class CalcManager(Manager):
    """
    Parent Class managing the true potential being simulated.
    This Calc class can support any :class:`ase.calculator` calculators from
    the ASE python package.
    Create, execute and read the output of any ASE supported calculators.

    Parameters
    ----------
    calc: :class:`ase.calculator`
        An ASE calculator object

    magmoms: :class:`np.ndarray` (optional)
        An array for the initial magnetic moments for each computation
        If ``None``, no initial magnetization. (Non magnetic calculation)
        Default ``None``.

    folder: :class:`str` (optional)
        The root for the directory in which the computation are to be done
        Default 'Calc'
    """

    def __init__(self,
                 calc,
                 magmoms=None,
                 folder='Calc',
                 **kwargs):

        Manager.__init__(self, folder=folder, **kwargs)

        self.calc = calc
        self.magmoms = magmoms

# ========================================================================== #
    def compute_true_potential(self, confs, subfolder=None, step=None):
        """
        Compute the energy/forces/stress of given input configurations with an
        ASE calculator.

        Parameters
        ----------
        confs: :class:`list` of :class:`ase.Atoms`
            The input list of atom objects.

        subfolder: :class:`list` of :class:`str` (optional)
            Subfolder in which the properties are saved.

        step: :class:`list` of :class:`int`  (optional)
            The list of configuration indices.

        Returns
        -------
        result_confs: :class:`list` of :class:`ase.Atoms`
            The output list of atom objects, with corresponding
            SinglePointCalculator resulting from true potential calculation.
        """
        confs = [at.copy() for at in confs]
        result_confs = []
        for at in confs:
            at.set_initial_magnetic_moments(self.magmoms)
            at.calc = self.calc
            try:
                energy = at.get_potential_energy()
                forces = at.get_forces()
                stress = at.get_stress()
                sp_calc = SinglePointCalculator(at,
                                                energy=energy,
                                                forces=forces,
                                                stress=stress)
                at.calc = sp_calc
                result_confs.append(at)
            except CalculatorError:
                result_confs.append(None)
        return result_confs

# ========================================================================== #
    def log_recap_state(self):
        """
        """
        try:
            name = self.calc.name
        except AttributeError:
            name = None

        msg = "True potential parameters:\n"
        if name is not None:
            msg += "Calculator : {0}\n".format(name)
        if hasattr(self.calc, "todict"):
            dct = self.calc.todict()
            msg += "parameters :\n"
            for key in dct.keys():
                msg += "   " + key + "  {0}\n".format(dct[key])
        msg += "\n"
        return msg



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import argparse
import textwrap
from argparse import RawTextHelpFormatter
from importlib import import_module

from ..version import __version__


commands = [('correlation', 'mlacs.cli.correlation'),
            ('plot_error', 'mlacs.cli.plot_error'),
            ('plot_weights', 'mlacs.cli.plot_weights'),
            ('plot_thermo', 'mlacs.cli.plot_thermo'),
            ('plot_neff', 'mlacs.cli.plot_neff')]


def main(prog='mlacs', description='MLACS command line tool',
         version=__version__, commands=commands, hook=None, args=None):
    parser = argparse.ArgumentParser(prog=prog,
                                     description=description,
                                     formatter_class=RawTextHelpFormatter)
    parser.add_argument('--version', action='version',
                        version=f"{prog}-{__version__}")
    subparsers = parser.add_subparsers(title="Sub-commands",
                                       dest="command")
    subparser = subparsers.add_parser('help',
                                      description='Help',
                                      help='Help for sub-command.')

    function = {}
    parsers = {}
    for command, module_name in commands:
        cmd = import_module(module_name).CLICommand
        docstring = cmd.__doc__
        parts = docstring.split("\n", 1)
        short, body = parts
        long = short + "\n" + textwrap.dedent(body)
        subparser = subparsers.add_parser(
            command,
            help=short,
            description=long,
            formatter_class=RawTextHelpFormatter)
        cmd.add_arguments(subparser)
        function[command] = cmd.run
        parsers[command] = subparser

    if hook:
        args = hook(parser, args)
    else:
        args = parser.parse_args(args)

    if args.command == "help":
        parser.print_help()
    elif args.command is None:
        parser.print_usage()
    else:
        f = function[args.command]
        try:
            if f.__code__.co_argcount == 1:
                f(args)
            else:
                f(args, parsers[args.command])
        except KeyboardInterrupt:
            pass



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np
from ase.atoms import Atoms
from ase.build import bulk

from ... import context  # noqa
from mlacs.utilities.miscellanous import (get_elements_Z_and_masses,
                                          create_random_structures)


def test_get_elements():
    """
    Here we test that the function get_elements_Z_and_masses() returns
    the unique elements, Z, masses and charges, ordered alphabetically
    with the elements
    """
    # First try with some charges
    elem = np.array(["Si", "Fe", "Rh", "U", "H", "H", "Si", "Fe"])
    charges = [1, 2, 0, 0, -1, -1, 1, 2]
    pos = np.zeros((len(elem), 3))
    cell = np.eye(3)

    idx = np.array([1, 4, 2, 0, 3])

    at = Atoms(elem, pos, cell=cell, charges=charges)
    ref_masses = at.get_masses()[idx]
    ref_charges = at.get_initial_charges()[idx]
    ref_elem = elem[idx]
    ref_Z = at.get_atomic_numbers()[idx]

    pred_el, pred_Z, pred_masses, pred_charges = get_elements_Z_and_masses(at)
    for ref, pred in zip(ref_elem, pred_el):
        assert ref == pred
    for ref, pred in zip(ref_Z, pred_Z):
        assert ref == pred
    for ref, pred in zip(ref_masses, pred_masses):
        assert ref == pred
    for ref, pred in zip(ref_charges, pred_charges):
        assert ref == pred

    # The same without charges
    elem = np.array(["Si", "Fe", "Rh", "U", "H", "H", "Si", "Fe"])
    pos = np.zeros((len(elem), 3))
    cell = np.eye(3)

    idx = np.array([1, 4, 2, 0, 3])

    at = Atoms(elem, pos, cell=cell)
    ref_masses = at.get_masses()[idx]
    ref_charges = at.get_initial_charges()[idx]
    ref_elem = elem[idx]
    ref_Z = at.get_atomic_numbers()[idx]

    pred_el, pred_Z, pred_masses, pred_charges = get_elements_Z_and_masses(at)
    for ref, pred in zip(ref_elem, pred_el):
        assert ref == pred
    for ref, pred in zip(ref_Z, pred_Z):
        assert ref == pred
    for ref, pred in zip(ref_masses, pred_masses):
        assert ref == pred
    assert pred_charges is None


def test_create_random_structures():
    """
    Here, we check that create_random_structures() return the right number
    of structures, and that the structures are different
    (a problem that can happend when atoms.rattle() is not used rightly)
    """
    at = bulk("Si").repeat(2)
    nconfs = 1
    newconfs = create_random_structures(at, std=0.01, nconfs=nconfs)
    assert len(newconfs) == nconfs
    nconfs = 4
    newconfs = create_random_structures(at, std=0.01, nconfs=nconfs)
    assert len(newconfs) == nconfs

    for at in newconfs[1:]:
        assert at != newconfs[0]



"""
// Copyright (C) 2022-2024 MLACS group (AC, CD)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest

from pathlib import Path
import shutil
import os


@pytest.fixture(autouse=True)
def root():
    return Path().absolute()


@pytest.fixture(autouse=True)
def clean_up_hist(root):
    str_dir = "tmp_hist_dir"
    dir_list = [x[0] for x in os.walk(root) if x[0][:-1].endswith(str_dir)]
    for directory in dir_list:
        shutil.rmtree(directory)
    yield
    dir_list = [x[0] for x in os.walk(root) if x[0][:-1].endswith(str_dir)]
    for directory in dir_list:
        shutil.rmtree(directory)



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np
import mlacs.utilities.eos_functions as eos_functions

from scipy.optimize import curve_fit

Bohr_GPa = 29421.033


def eos_fit(x=None, y=None, y_type='presssure', eos='vinet', **kwargs):
    '''
    fit an eos from E(V) or P(V) on an isotherm
    return fitted isothermal params (e0,) b0, b0p, v0

    be aware of used units:
    - for P(V) GPa and Angstrom^3 work
    - for E(V) all should be in Ha (Bohr^3, Ha/Bohr^3)
      or in eV (Angstrom^3 and eV/Angstrom^3)

    Parameters
    ----------

    x: volume 1D array,
       per atom pay an attention to hexa phases for instance
    y: energy or pressure 1D array
    y_type: str
        pressure or volume
    model: str
        birch-murnaghan
        murnaghan
        vinet

    Optionnal
    ---------

    plot:
    save:
    error:
    '''

    # get x and y from canonical sampling if not given
    # initial guess
    a, b, c = np.polyfit(x, y, 2)
    v0 = -b/(2*a)
    b0 = 2*a*v0
    b0p = 4.0

    if y_type == 'energy':
        e0 = a * v0**2 + b * v0 + c
        bounds = ([e0 * 2, b0 / 100, b0p / 10, v0 / 2],
                  [e0 / 2, b0 * 100, b0p * 10, v0 * 2])
        if eos == 'vinet':
            fitted_params, ecov_ = curve_fit(eos_functions.e_vinet, x, y,
                                             bounds=bounds)
        if eos == 'murnaghan':
            fitted_params, ecov_ = curve_fit(eos_functions.e_murnaghan, x, y,
                                             bounds=bounds)
        if eos == 'bm':
            fitted_params, ecov_ = curve_fit(eos_functions.e_bm, x, y,
                                             bounds=bounds)

        str = 'Fitted Parameters from EoS fit\n'
        str += '------------------------------\n'
        str += 'V0 = {:12.8f} Angstrom^3\n'.format(fitted_params[3])
        str += 'B0 = {:12.8f} GPa\n'.format(fitted_params[1])
        str += 'B0p = {:12.8f} \n'.format(fitted_params[2])
        str += 'E0 = {:12.8f} eV/at\n'.format(fitted_params[0])

    else:
        bounds = ([b0 / 100, b0p / 10, v0 / 2], [b0 * 100, b0p * 10, v0 * 2])
        # ph = 1
        if eos == 'vinet':
            fitted_params, pcov_ = curve_fit(eos_functions.p_vinet, x, y,
                                             bounds=bounds)
        if eos == 'murnaghan':
            fitted_params, pcov_ = curve_fit(eos_functions.p_murnaghan, x, y,
                                             bounds=bounds)
        if eos == 'bm':
            fitted_params, pcov_ = curve_fit(eos_functions.p_bm, x, y,
                                             bounds=bounds)

        str = 'Fitted Parameters from EoS fit\n'
        str += '------------------------------\n'
        str += 'V0 = {:12.8f} Angstrom^3\n'.format(fitted_params[2])
        str += 'B0 = {:12.8f} GPa\n'.format(fitted_params[0])
        str += 'B0p = {:12.8f}\n'.format(fitted_params[1])

    print(str)
    return fitted_params



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from pathlib import Path
from subprocess import run, Popen, PIPE
from os import symlink
import shutil
import shlex

import numpy as np
from ase.calculators.lammpsrun import LAMMPS
from ase.units import GPa

from .descriptor import BlankDescriptor
from .mlip_manager import SelfMlipManager
from ..utilities import compute_correlation
from ..utilities.io import write_cfg, read_cfg_data


default_mtp_parameters = dict(level=8,
                              radial_basis_type="RBChebyshev",
                              min_dist=1.0,
                              max_dist=5.0,
                              radial_basis_size=8)

default_fit_parameters = dict(scale_by_forces=0,
                              max_iter=100,
                              bfgs_conv_tol=1e-3,
                              weighting="vibrations",
                              weight_scaling=None,
                              weight_scaling_forces=None,
                              init_params="random",
                              update_mindist=True)


# ========================================================================== #
# ========================================================================== #
class MomentTensorPotential(SelfMlipManager):
    """
    Interface to the Moment Tensor Potential of the MLIP package.

    Parameters
    ----------
    atoms: :class:`ase.Atoms`
        Prototypical configuration for the MLIP. Should have the desired
        species.

    mlpbin: :class:`str`
        The path to the  `mlp` binary. If mpi is desired, the command
        should be set as 'mpirun /path/to/mlp'

    mtp_parameters: :class:`dict`
        The dictionnary with inputs for the potential.

        The default values are set to

        - level = 8
        - radial_basis_type = 'RBChebyshev'
        - min_dist=1.0,
        - max_dist=5.0,
        - radial_basis_size=8

    fit_parameters: :class:`dict`
        The parameters for the fit of the potential

        The default parameters are set to
            - scale_by_forces=0
            - max_iter=1000
            - bfgs_conv_tol=1e-3
            - weighting='vibrations'
            - (weight_scaling=1, weight_scaling_forces=0 for MLIP-3)
            - init_params='random'
            - update_mindist=False

    Examples
    --------

    >>> from ase.io import read
    >>> confs = read('Trajectory.traj', index=':')
    >>>
    >>> from mlacs.mlip import MomentTensorPotential
    >>> mlip = MomentTensorPotential(confs[0], mtp_parameters=dict(level=6))
    >>> mlip.update_matrices(confs)
    >>> mlip.train_mlip()
    """
    def __init__(self,
                 atoms,
                 mlpbin="mlp",
                 mtp_parameters={},
                 fit_parameters={},
                 folder='MTP',
                 **kwargs):

        self.cmd = mlpbin
        self.version = 2
        testcmd = shlex.split(f'{mlpbin} list')
        mlp_info = Popen(testcmd, stdout=PIPE).communicate()[0]
        if b'calculate_efs' in mlp_info:
            self.version = 3

        SelfMlipManager.__init__(self,
                                 descriptor=BlankDescriptor(atoms),
                                 weight=None,
                                 folder=folder,
                                 **kwargs)

        self.level = mtp_parameters.pop("level", 8)
        if self.level % 2 or self.level > 28:
            msg = "Only even number between 2 and 28 are available as level"
            raise ValueError(msg)
        self.mtp_parameters = default_mtp_parameters
        for key in mtp_parameters.keys():
            if key not in list(default_mtp_parameters.keys()):
                msg = f"{key} is not a parameter for the MTP potential"
                raise ValueError(msg)
        self.mtp_parameters.update(mtp_parameters)

        self.fit_parameters = default_fit_parameters
        self.fit_parameters.update(fit_parameters)

# ========================================================================== #
    def read_parent_mlip(self, traj):
        """
        Get a list of all the mlip that have generated a conf in traj
        and get the coefficients of all these mlip
        """
        parent_mlip = []
        mlip_coef = []
        # Make the MBAR variable Nk and mlip_coef
        for state in traj:
            for conf in state:
                if "parent_mlip" not in conf.info:  # Initial or training
                    continue
                else:  # A traj
                    model = conf.info['parent_mlip']
                    if not Path(model).exists:
                        err = "Some parent MLIP are missing. "
                        err += "Rerun MLACS with DatabaseCalculator and "
                        err += "OtfMlacs.keep_tmp_files=True on your traj"
                        raise FileNotFoundError(err)
                    if model not in parent_mlip:  # New state
                        parent_mlip.append(model)
                        coef = model
                        mlip_coef.append(coef)
        return parent_mlip, np.array(mlip_coef)

# ========================================================================== #
    def next_coefs(self, mlip_coef):
        """
        Update MLACS just like train_mlip, but without actually computing
        the coefficients
        """
        pass

# ========================================================================== #
    def get_pair_style(self):
        if self.version == 3:
            return f"mlip load_from={self.subdir / 'pot.mtp'}"
        return f"mlip {self.subdir / 'mlip.ini'}"

# ========================================================================== #
    def get_pair_coeff(self):
        return ["* *"]

# ========================================================================== #
    def get_pair_style_coeff(self):
        return self.get_pair_style(), \
               self.get_pair_coeff()

# ========================================================================== #
    def train_mlip(self):
        """
        """
        # We need to remove the old subfolder. If calculation ended at
        # some specific weird moments, it can cause access problem
        # for the mlp binary otherwise
        if self.subfolder:
            if self.subsubdir.exists():
                shutil.rmtree(self.subsubdir)

        self.subsubdir.mkdir(parents=True, exist_ok=True)

        mtpfile = self.subdir / "pot.mtp"

        # Move old potential in new folder, to start BFGS from
        # previously trained MTP
        if mtpfile.exists():
            shutil.move(mtpfile, self.subsubdir / "pot.mtp")

        self._clean_folder()
        self._write_configurations()
        self._write_input()
        self._write_mtpfile()
        self._run_mlp()

        # Symlink new MTP in the main folder
        if self.subfolder:
            # We might need to remove the old symlink before the new one
            if mtpfile.exists():
                mtpfile.unlink()

            src = self.subsubdir / "pot.mtp"
            symlink(src, self.subdir / "pot.mtp")

        with open(self.subdir / "mlip.ini", "w") as fd:
            fd.write(f"mtp-filename    {mtpfile}\n")
            fd.write("select           FALSE")

        msg = "number of configurations for training: " + \
              f"{len(self.natoms):}\n"
        msg += "number of atomic environments for training: " + \
               f"{self.natoms.sum():}\n"
        msg += self._compute_test(msg)

        if self.log:
            self.log.write(msg)

# ========================================================================== #
    def get_calculator(self):
        """
        Initialize a ASE calculator from the model
        """
        calc = LAMMPS(pair_style=self.pair_style,
                      pair_coeff=self.pair_coeff,
                      atom_style=self.atom_style,
                      keep_alive=False)
        if self.model_post is not None:
            calc.set(model_post=self.model_post)
        return calc

# ========================================================================== #
    def _clean_folder(self):
        """
        """
        files = ["train.cfg",
                 "mlip.ini",
                 "initpot.mtp",
                 "out.cfg"]
        for fn in files:
            if (filepath := self.subsubdir / fn).exists():
                filepath.unlink()

# ========================================================================== #
    def _write_configurations(self):
        """
        """
        confs = self.configurations
        chemmap = self.descriptor.elements
        write_cfg(self.subsubdir / "train.cfg", confs, chemmap)

# ========================================================================== #
    def _write_input(self):
        """
        """
        mtpfile = self.subsubdir / "pot.mtp"
        with open(self.subsubdir / "mlip.ini", "w") as fd:
            fd.write(f"mtp-filename    {mtpfile}\n")
            fd.write("select           FALSE")

# ========================================================================== #
    def _write_mtpfile(self):
        """
        """
        writenewmtp = True
        mtpfile = self.subsubdir / "initpot.mtp"
        lvl = self.level
        level = f"level{lvl}"
        if (potfile := self.subsubdir / "pot.mtp").exists():
            import re
            with open(potfile, "r", encoding="ISO-8859-1") as fd:
                for line in fd.readlines():
                    if line.startswith("potential_name"):
                        oldlevel = int(re.search(r'\d+$', line).group())
                        break
            if oldlevel == lvl:
                potfile.rename(mtpfile)
                writenewmtp = False
        if writenewmtp:
            from . import _mtp_data
            leveltxt = getattr(_mtp_data, level)
            nel = self.descriptor.nel
            btype = self.mtp_parameters["radial_basis_type"]
            min_dist = self.mtp_parameters["min_dist"]
            max_dist = self.mtp_parameters["max_dist"]
            bsize = self.mtp_parameters["radial_basis_size"]
            with open(mtpfile, "w") as fd:
                fd.write("MTP\n")
                fd.write("version = 1.1.0\n")
                fd.write(f"potential_name = MTP-{level}\n")
                fd.write(f"species_count = {nel}\n")
                fd.write("potential_tag = \n")
                fd.write(f"radial_basis_type = {btype}\n")
                fd.write(f"min_dist = {min_dist}\n")
                fd.write(f"max_dist = {max_dist}\n")
                fd.write(f"radial_basis_size = {bsize}\n")
                fd.write(leveltxt)

# ========================================================================== #
    def _run_mlp(self):
        """
        """
        if self.version == 3:
            mlp_command = self._get_cmd_mlip3()
        else:
            mlp_command = self._get_cmd_mlip2()
        with open(self.subsubdir / "mlip.log", "w") as fd:
            mlp_handle = run(mlp_command.split(),
                             stderr=PIPE,
                             stdout=fd,
                             cwd=self.subsubdir)
        if mlp_handle.returncode != 0:
            msg = "mlp stopped with the exit code \n" + \
                  f"{mlp_handle.stderr.decode()}"
            raise RuntimeError(msg)

# ========================================================================== #
    def _get_cmd_mlip2(self):
        """
        Get mlp command from MLIP-2 package.
        """
        initpotfile = self.subsubdir / "initpot.mtp"
        potfile = self.subsubdir / "pot.mtp"
        trainfile = self.subsubdir / "train.cfg"
        cmd = self.cmd + f" train {initpotfile} {trainfile}"
        cmd += f" --trained-pot-name={potfile}"
        up_mindist = self.fit_parameters["update_mindist"]
        if up_mindist:
            cmd += " --update-mindist"
        init_params = self.fit_parameters["init_params"]
        cmd += f" --init-params={init_params}"
        max_iter = self.fit_parameters["max_iter"]
        cmd += f" --max-iter={max_iter}"
        bfgs_conv_tol = self.fit_parameters["bfgs_conv_tol"]
        cmd += f" --bfgs-conv-tol={bfgs_conv_tol}"
        scale_by_forces = self.fit_parameters["scale_by_forces"]
        cmd += f" --scale-by-force={scale_by_forces}"
        cmd += f" --energy-weight={self.weight.energy_coefficient}"
        cmd += f" --force-weight={self.weight.forces_coefficient}"
        cmd += f" --stress-weight={self.weight.stress_coefficient}"
        weighting = self.fit_parameters["weighting"]
        cmd += f" --weighting={weighting}"
        return cmd

# ========================================================================== #
    def _get_cmd_mlip3(self):
        """
        Get mlp command from MLIP-3 package.
        """
        initpotfile = self.subsubdir / "initpot.mtp"
        potfile = self.subsubdir / "pot.mtp"
        trainfile = self.subsubdir / "train.cfg"
        cmd = self.cmd + f" train {initpotfile} {trainfile}"
        cmd += f" --save_to={potfile}"
        # Default in MLIP-3 is mindist update, which is the opposite of MLIP-2.
        # Need to check this ...
        up_mindist = not self.fit_parameters["update_mindist"]
        cmd += f" --no_mindist_update={up_mindist}"
        if self.fit_parameters["init_params"] == "random":
            cmd += " --init_random=True"
        max_iter = self.fit_parameters["max_iter"]
        cmd += f" --iteration_limit={max_iter}"
        bfgs_conv_tol = self.fit_parameters["bfgs_conv_tol"]
        cmd += f" --tolerance={bfgs_conv_tol}"
        scale_by_forces = self.fit_parameters["scale_by_forces"]
        cmd += f" --scale_by_force={scale_by_forces}"
        cmd += f" --energy_weight={self.weight.energy_coefficient}"
        cmd += f" --force_weight={self.weight.forces_coefficient}"
        cmd += f" --stress_weight={self.weight.stress_coefficient}"
        w_scale = self.fit_parameters["weight_scaling"]
        w_f_scale = self.fit_parameters["weight_scaling_forces"]
        if w_scale is not None or w_f_scale is not None:
            if w_scale is not None:
                cmd += f" --weight_scaling={w_scale}"
            if w_f_scale is not None:
                cmd += f" --weight_scaling_forces={w_f_scale}"
            return cmd
        weighting = self.fit_parameters["weighting"]
        if weighting == "vibrations":
            cmd += " --weight_scaling=1 --weight_scaling_forces=0"
        elif weighting == "molecules":
            cmd += " --weight_scaling=0 --weight_scaling_forces=0"
        elif weighting == "structures":
            cmd += " --weight_scaling=2 --weight_scaling_forces=1"
        return cmd

# ========================================================================== #
    def _get_pair_style(self):
        return self.get_pair_style()

# ========================================================================== #
    def _get_pair_coeff(self):
        return self.get_pair_coeff()

# ========================================================================== #
    def predict(self, atoms, coef=None):
        msg = "Getting the MLIP energy from descriptor is not accessible " + \
              "for SelfMlipManager"
        raise NotImplementedError(msg)

# ========================================================================== #
    def _run_test(self):
        """
        """
        trainfile = self.subsubdir / "train.cfg"
        outfile = self.subsubdir / "out.cfg"
        potfile = self.subsubdir / "pot.mtp"
        mlp_command = self.cmd
        if self.version == 3:
            mlp_command += f" calculate_efs {potfile} {trainfile}"
            mlp_command += f" --output_filename={outfile}"
            outfile = self.subsubdir / "out.cfg.0"
        else:
            mlp_command += f" calc-efs {potfile} {trainfile} {outfile}"
        mlp_handle = run(mlp_command.split(),
                         stderr=PIPE)
        if mlp_handle.returncode != 0:
            msg = "mlp stopped with the exit code \n" + \
                  f"{mlp_handle.stderr.decode()}"
            raise RuntimeError(msg)
        e_mlip, f_mlip, s_mlip = read_cfg_data(outfile)
        return e_mlip, f_mlip, s_mlip

# ========================================================================== #
    def _compute_test(self, msg):
        """
        """
        e_mlip, f_mlip, s_mlip = self._run_test()

        confs = self.configurations
        e_dft = np.array([at.get_potential_energy() / len(at)for at in confs])
        f_dft = []
        s_dft = []
        for at in confs:
            f_dft.extend(at.get_forces().flatten())
            s_dft.extend(at.get_stress())

        rmse_e, mae_e, rsq_e = compute_correlation(np.c_[e_dft, e_mlip])
        rmse_f, mae_f, rsq_f = compute_correlation(np.c_[f_dft, f_mlip])
        rmse_s, mae_s, rsq_s = compute_correlation(np.c_[s_dft, s_mlip] / GPa)

        # Prepare message to the log
        msg += f"RMSE Energy    {rmse_e:.4f} eV/at\n"
        msg += f"MAE Energy     {mae_e:.4f} eV/at\n"
        msg += f"RMSE Forces    {rmse_f:.4f} eV/angs\n"
        msg += f"MAE Forces     {mae_f:.4f} eV/angs\n"
        msg += f"RMSE Stress    {rmse_s:.4f} GPa\n"
        msg += f"MAE Stress     {mae_s:.4f} GPa\n"
        msg += "\n"

        header = f"rmse: {rmse_e:.5f} eV/at,    " + \
                 f"mae: {mae_e:.5f} eV/at\n" + \
                 " True Energy           Predicted Energy"
        np.savetxt("MLIP-Energy_comparison.dat",
                   np.c_[e_dft, e_mlip],
                   header=header, fmt="%25.20f  %25.20f")
        header = f"rmse: {rmse_f:.5f} eV/angs   " + \
                 f"mae: {mae_f:.5f} eV/angs\n" + \
                 " True Forces           Predicted Forces"
        np.savetxt("MLIP-Forces_comparison.dat",
                   np.c_[f_dft, f_mlip],
                   header=header, fmt="%25.20f  %25.20f")
        header = f"rmse: {rmse_s:.5f} GPa       " + \
                 f"mae: {mae_s:.5f} GPa\n" + \
                 " True Stress           Predicted Stress"
        np.savetxt("MLIP-Stress_comparison.dat",
                   np.c_[s_dft, s_mlip] / GPa,
                   header=header, fmt="%25.20f  %25.20f")
        return msg

# ========================================================================== #
    def __str__(self):
        txt = f"Moment Tensor Potential, level = {self.level}"
        return txt

# ========================================================================== #
    def __repr__(self):
        txt = "Moment Tensor Potential\n"
        txt += "Parameters:\n"
        txt += "-----------\n"
        txt += "Descriptor:\n"
        txt += "-----------\n"
        txt += f"level :                 {self.level}\n"
        basis = self.mtp_parameters["radial_basis_type"]
        nbasis = self.mtp_parameters["radial_basis_size"]
        min_dist = self.mtp_parameters["min_dist"]
        max_dist = self.mtp_parameters["max_dist"]
        txt += f"radial basis function : {basis}\n"
        txt += f"Radial basis size :     {nbasis}\n"
        txt += f"Minimum distance :      {min_dist}\n"
        txt += f"Cutoff :                {max_dist}\n"
        return txt



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest
from pathlib import Path

from ase.build import bulk
from ase.calculators.emt import EMT

from ... import context  # noqa
from mlacs.mlip import MomentTensorPotential
from mlacs.state import LammpsState
from mlacs import OtfMlacs


"""
Example of a MLACS simulation of Cu at several temperature
The true potential is the EMT as implemented in ASE
"""


@pytest.fixture
def expected_folder(expected_folder_base):
    return ["MTP"]


@pytest.fixture
def expected_files(expected_files_base):
    return expected_files_base


@pytest.mark.skipif(context.has_mlp(), reason="Need MLIP-2 to use MTP")
def test_mtp_restart(root, treelink):
    # Parameters --------------------------------------------------------------
    temperature = 300  # K
    nconfs = 2
    nsteps = 10
    nsteps_eq = 10
    neq = 1
    cell_size = 2
    dt = 1.5  # fs
    mtp_params = dict(level=8,
                      radial_basis_type="RBChebyshev",
                      min_dist=1.0,
                      max_dist=5.0,
                      radial_basis_size=8)

    mtp_fit = dict(scale_by_forces=0,
                   max_iter=100,
                   bfgs_conv_tol=1e-3,
                   weighting="vibrations",
                   init_params="random",
                   update_mindist=False)

    # Supercell creation ------------------------------------------------------
    atoms = bulk('Cu', cubic=True).repeat(cell_size)
    calc = EMT()

    # Running Initial Lammps State
    state = LammpsState(temperature, dt=dt, nsteps=nsteps,
                        nsteps_eq=nsteps_eq)

    mtp = MomentTensorPotential(atoms,
                                mlpbin="mlp",
                                folder=Path("MTP").absolute(),
                                mtp_parameters=mtp_params,
                                fit_parameters=mtp_fit)

    sampling = OtfMlacs(atoms, state, calc, mtp, neq=neq)
    sampling.run(nconfs)

    # Running Restart Lammps State
    state = LammpsState(temperature, dt=dt, nsteps=nsteps,
                        nsteps_eq=nsteps_eq)

    mtp = MomentTensorPotential(atoms,
                                mlpbin="mlp",
                                folder=Path("MTP").absolute(),
                                mtp_parameters={},
                                fit_parameters={})

    sampling = OtfMlacs(atoms, state, calc, mtp, neq=neq)
    sampling.run(nconfs)
    assert sum(sampling.nconfs) == 4



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest

import numpy as np
from ase.build import bulk
from ase.io import read
from ase.calculators.emt import EMT

from ... import context  # noqa
from mlacs.state import OptimizeLammpsState
from mlacs.mlip import SnapDescriptor, LinearPotential, DeltaLearningPotential
from mlacs import MlMinimizer


@pytest.fixture
def files_with_prefix():
    files = []
    ptype = ['iso', 'iso', 'iso', 'aniso', 'aniso']
    press = [None, None, 0.0, 0.0, 5.0]
    algo = ['cg', 'fire', 'cg', 'cg', 'cg']
    for t, p, a in zip(ptype, press, algo):
        files.append(f'{a}_{p}_{t}.traj')
        files.append(f'{a}_{p}_{t}_potential.dat')
    return files


@pytest.fixture
def expected_folder():
    folder = ["MolecularDynamics", "Snap"]
    return folder


@pytest.fixture
def expected_files(files_with_prefix):
    files = ["MLACS.log", "MLACS.log0001", "MLACS.log0002", "MLACS.log0003",
             "MLACS.log0004", "Training_configurations.traj",
             "MLIP-Energy_comparison.dat", "MLIP-Forces_comparison.dat",
             "MLIP-Stress_comparison.dat"]
    files.extend(files_with_prefix)
    return files


@pytest.mark.slow
def test_mlacs_optimize(root, treelink):

    atoms = bulk("Cu", cubic=True).repeat(2)
    atoms.pop(0)
    natoms = len(atoms)
    nstep = 10
    calc = EMT()

    mlip_params = dict(twojmax=4)
    desc = SnapDescriptor(atoms, 4.2, mlip_params)
    mlip = LinearPotential(desc, folder="Snap")
    ps, cs = 'zbl 1.0 2.0', ['* * 29 29']
    dmlip = DeltaLearningPotential(mlip, pair_style=ps, pair_coeff=cs)

    etol = 0.01
    ftol = 0.01
    stol = 0.01

    prefix = []
    ptype = ['iso', 'iso', 'iso', 'aniso', 'aniso']
    press = [None, None, 0.0, 0.0, 5.0]
    algo = ['cg', 'fire', 'cg', 'cg', 'cg']
    for t, p, a in zip(ptype, press, algo):
        prefix.append(f'{a}_{p}_{t}')
        state = OptimizeLammpsState(min_style=a, pressure=p, ptype=t,
                                    prefix=prefix[-1])
        sampling = MlMinimizer(atoms, state, calc, dmlip, etol, ftol, stol)
        sampling.run(nstep)

    for folder in treelink["folder"]:
        assert (root / folder).exists()

    for file in treelink["files"]:
        assert (root / file).exists()

    for p in prefix:
        traj = read(root / f"{p}.traj", ":")
        # Check that the first atom is the one we started with
        assert traj[0] == atoms
        # Check that the system didn't change in the process
        for at in traj:
            assert len(at) == natoms
        # Check that the criterion on forces is achieved
        assert np.max(traj[-1].get_forces()) <= ftol
        # Check that volume is constant
        if 'None' in p:
            assert traj[-1].get_volume() == atoms.get_volume()



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest

import numpy as np

from ase.build import bulk

from ... import context  # noqa

from mlacs.calc import DatabaseCalc
from mlacs.mlip import MliapDescriptor, LinearPotential, MbarManager
from mlacs.state import LammpsState
from mlacs import OtfMlacs


"""
Test of DatabaseCalc and of Non-Regression for Mbar
"""


@pytest.fixture
def expected_folder(expected_folder_base):
    return ["Mliap", "MolecularDynamics"]


@pytest.fixture
def expected_files(expected_files_base):
    return expected_files_base


@pytest.mark.full
def test_mbar_databasecalc(root, treelink):
    ref = root.absolute() / "reference_files"
    # Parameters --------------------------------------------------------------
    temperature = 300  # K
    nconfs = 5
    nsteps = 10
    nsteps_eq = 10
    neq = 1
    cell_size = 2
    rcut = 4.2
    dt = 1.5  # fs
    mlip_params = {"twojmax": 4}

    # Supercell creation ------------------------------------------------------
    atoms = bulk('Cu', cubic=True).repeat(cell_size)

    # Weight
    mbar_params = dict(mode="train", solver="L-BFGS-B")
    mbar = MbarManager(parameters=mbar_params)

    # Running DatabaseCalc
    db_calc = DatabaseCalc(trajfile=ref / "Database.traj",
                           trainfile=ref / "Training_Database.traj")
    state = LammpsState(temperature, dt=dt, nsteps=nsteps, nsteps_eq=nsteps_eq)
    desc = MliapDescriptor(atoms, rcut, mlip_params, style="snap")
    mlip = LinearPotential(desc, folder="Mliap", weight=mbar)
    sampling = OtfMlacs(atoms, state, db_calc, mlip, neq=neq)
    sampling.run(nconfs)

    # Make sure the weight computation did not change
    precomputed_weight = [0.24974929, 0.13717322, 0.15500564,
                          0.20601895, 0.2520529]
    w = sampling.mlip.weight.weight
    assert np.allclose(precomputed_weight, w, atol=1e-5), \
        "It seems mbar evaluate weight differently than before."



"""
// Copyright (C) 2022-2024 MLACS group (AC, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from pathlib import Path
import ase
from .calc_manager import CalcManager


# ========================================================================== #
# ========================================================================== #
class DatabaseCalc(CalcManager):
    """
    Calculators that sequentially reads a previously calculated traj files.
    Normal utilisator want to set OtfMlacs.nstep to len(traj).
    Can be used with restart to append trajfile to the current traj.

    Parameters
    ----------
    calc: :class:`ase.calculator`
        A ASE calculator object

    trajfile: :class:`str` or :class:`pathlib.Path`
        The trajectory file from which DatabaseCalc will read

    trainfile: :class:`str`, :class:`pathlib.Path`,
        The training.traj file, configuration used for fitting but
        not counted for thermodynamic properties

    magmoms: :class:`np.ndarray` (optional)
        An array for the initial magnetic moments for each computation
        If ``None``, no initial magnetization. (Non magnetic calculation)
        Default ``None``.
    """
    def __init__(self,
                 trajfile,
                 trainfile,
                 magmoms=None,
                 **kwargs):
        CalcManager.__init__(self, "dummy", magmoms, **kwargs)
        if isinstance(trajfile, str) or isinstance(trajfile, Path):
            self.traj = [ase.io.read(trajfile, index=":")]
        else:
            self.traj = [ase.io.read(t, index=":") for t in trajfile]
        self.training = ase.io.read(trainfile, index=":")
        self.current_conf = 0

# ========================================================================== #
    def compute_true_potential(self,
                               mlip_confs,
                               state=None,
                               step=None):
        """
        Return the energy of given configurations contained in the Trajectory
        file. This is a way to replay a previously done MLACS simulations.
        """
        # 1. Create a copy of the next atoms in traj as true_atoms
        # 2. Modify mlip_atoms positions to match what we have in the traj
        # 3. Change the Parent MLIP that generated true_confs
        length_trajs = sum([len(t) for t in self.traj])
        assert len(mlip_confs) + self.current_conf <= length_trajs, \
            "You cannot do more step than there is in the Trajectory file" +\
            "\nNumber of conf in the given Trajectory. If multiple state" +\
            " you need to give a list of Traj file."

        true_confs = []
        i = 0
        for mlip_conf, s in zip(mlip_confs, state):
            if s == "Training":
                true_confs.append(self.training.pop())
                continue

            true_confs.append(self.traj[i][self.current_conf])
            mlip_conf.set_positions(true_confs[-1].get_positions())
            true_confs[-1].info = mlip_conf.info
            i += 1
        if state[0] != "Training":
            self.current_conf += 1

        return true_confs



"""
// Copyright (C) 2022-2024 MLACS group (AC, CD)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt
from ..utilities.plots import plot_weights, init_rcParams, HistPlot


def main(args, parser):
    if '_HIST.nc' in args.file:
        ncname = args.file
        path = Path().absolute()
        ncpath = str(path / ncname)
        ncplot = HistPlot(ncpath=ncpath)
        boolean_no_show = args.noshow
        boolean_show = not boolean_no_show
        if args.iteration is not None:
            args.iteration = [int(x) for x in args.iteration]

        ncplot.plot_ith_weights(show=boolean_show, savename=args.save,
                                selected_iter=args.iteration)

    else:
        weights = np.loadtxt(args.file)

        fig = plt.figure(figsize=(20, 10), constrained_layout=True)
        init_rcParams()
        ax = fig.add_subplot()

        plot_weights(ax, weights)
        if args.save is not None:
            plt.savefig(args.save)
        if not args.noshow:
            plt.show()


class CLICommand:
    """
    Plot barplot of the weights. The file from which the data is loaded can be
    either
        MLIP.weights (only last MLACS iteration),
        *HIST.nc file (any MLACS iteration can be given, or a list of MLACS
                       iteration; if no iteration is given, then the last one
                       if shown by default).

    Examples:

        $ mlacs plot_weights MLIP.weights
        $ mlacs plot_weights example_HIST.nc
        $ mlacs plot_weights example_HIST.nc -i 2 5 10
    """
    @staticmethod
    def add_arguments(parser):
        parser.add_argument('file', help="file with the data")
        parser.add_argument("-s", "--save", default=None,
                            help="Name of the file to save the plot. "
                                 "Default None")
        parser.add_argument("--noshow", action="store_true",
                            help="To disable the visualisation of the plot")
        parser.add_argument('-i', '--iteration', nargs='+', default=None,
                            help="List of MLACS iteration index ")

    @staticmethod
    def run(args, parser):
        main(args, parser)



"""
// Copyright (C) 2022-2024 MLACS group (AC, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import shutil
import numpy as np
from ase import Atoms
from . import TensorpotPotential, MomentTensorPotential, MbarManager
from . import LinearPotential, DeltaLearningPotential
from .weighting_policy import WeightingPolicy


def split_dataset(confs, train_ratio=0.5, rng=None):
    """
    Split the dataset into a train set and a test set.
    """
    if rng is None:
        rng = np.random.default_rng()
    nconfs = len(confs)

    ntrain = int(np.ceil(nconfs * train_ratio))

    allidx = np.arange(0, nconfs)
    trainset_idx = rng.choice(allidx, ntrain, replace=False)
    testset_idx = list(set(allidx) - set(trainset_idx))

    trainset = []
    for i in trainset_idx:
        trainset.append(confs[i])
    testset = []
    for i in testset_idx:
        testset.append(confs[i])

    return trainset, testset


def linfit_traj(traj, mlip):
    """
    Fit an MLIP according to the trajectory
    """
    if not isinstance(traj[0], Atoms):
        raise ValueError("Traj must be an Ase.Trajectory")
    if not (isinstance(mlip, LinearPotential)
            or isinstance(mlip, DeltaLearningPotential)):
        msg = "Only LinearPotential or DeltaLearningPotential " + \
              "are allowed for linfit_traj"
        raise NotImplementedError(msg)
    atoms = [at for at in traj]
    if mlip.subfolder:
        if mlip.subsubdir.exists():
            shutil.rmtree(mlip.subsubdir)
    mlip.subsubdir.mkdir(parents=True, exist_ok=True)

    mlip.update_matrices(atoms)
    mlip.train_mlip()


def mtpfit_traj(traj, mlip):
    """
    Fit an MLIP according to the trajectory
    """
    if not isinstance(traj[0], Atoms):
        raise ValueError("Traj must be an Ase.Trajectory")
    if not isinstance(mlip, MomentTensorPotential):
        msg = "Only MomentTensorPotential are allowed for mtpfit_traj"
        raise NotImplementedError(msg)
    atoms = [at for at in traj]
    if mlip.subfolder:
        if mlip.subsubdir.exists():
            shutil.rmtree(mlip.subsubdir)
    mlip.subsubdir.mkdir(parents=True, exist_ok=True)

    mlip.update_matrices(atoms)
    mlip._write_configurations()
    mlip._write_input()
    mlip._write_mtpfile()
    mlip._run_mlp()


def acefit_traj(traj, mlip, weights=None, initial_potential=None):
    """
    Fit an MLIP according to the trajectory
    initial_potential :
        Potential to start the fitting from.
        Useful to reconverge with a better precision on a parameter.
        Can be a filename (str) or a BBasisConfiguration
    """
    if isinstance(weights, list):
        weights = np.array(weights)
    if not isinstance(traj[0], Atoms):
        raise ValueError("Traj must be an Ase.Trajectory")
    if not isinstance(mlip, TensorpotPotential):
        msg = "Only Tensorpotential are allowed for acefit_traj"
        raise NotImplementedError(msg)

    # Prepare the data
    atoms = [at for at in traj]
    mlip.update_matrices(atoms)

    if weights is None:
        if isinstance(mlip.weight, MbarManager):
            msg = "Use another WeightingPolicy in the mlip or give weight."
            raise ValueError(msg)
        mlip.weight.compute_weight()
        weights = mlip.weight.get_weights()
    else:
        if len(atoms) == len(weights):
            we, wf, ws = WeightingPolicy(database=atoms).build_W_efs(weights)
            weights = np.append(np.append(we, wf), ws)
    if initial_potential is not None:
        mlip.descriptor.bconf.load(initial_potential)
        curr_fc = mlip.descriptor.bconf.metadata['_fit_cycles']
        mlip.descriptor.fitting['fit_cycles'] = int(curr_fc) + 1
    mlip.descriptor.redirect_logger()
    mlip.descriptor.fit(weights=weights, atoms=atoms)
